@@ -15,16 +15,31 @@
 #include "miscadmin.h"
 
 #include "access/xact.h"
+#include "distributed/backend_data.h"
 #include "distributed/connection_management.h"
 #include "distributed/metadata_cache.h"
 #include "distributed/remote_commands.h"
 #include "distributed/remote_transaction.h"
+#include "distributed/transaction_identifier.h"
 #include "distributed/transaction_management.h"
 #include "distributed/transaction_recovery.h"
 #include "distributed/worker_manager.h"
 #include "utils/hsearch.h"
 
 
+static void StartRemoteTransactionSavepointBegin(MultiConnection *connection,
+												 SubTransactionId subId);
+static void FinishRemoteTransactionSavepointBegin(MultiConnection *connection,
+												  SubTransactionId subId);
+static void StartRemoteTransactionSavepointRelease(MultiConnection *connection,
+												   SubTransactionId subId);
+static void FinishRemoteTransactionSavepointRelease(MultiConnection *connection,
+													SubTransactionId subId);
+static void StartRemoteTransactionSavepointRollback(MultiConnection *connection,
+													SubTransactionId subId);
+static void FinishRemoteTransactionSavepointRollback(MultiConnection *connection,
+													 SubTransactionId subId);
+
 static void CheckTransactionHealth(void);
 static void Assign2PCIdentifier(MultiConnection *connection);
 static void WarnAboutLeakedPreparedTransaction(MultiConnection *connection, bool commit);
@@ -32,12 +47,18 @@ static void WarnAboutLeakedPreparedTransaction(MultiConnection *connection, bool
 
 /*
  * StartRemoteTransactionBeging initiates beginning the remote transaction in
- * a non-blocking manner.
+ * a non-blocking manner. The function sends "BEGIN" followed by
+ * assign_distributed_transaction_id() to assign the distributed transaction
+ * id on the remote node.
  */
 void
 StartRemoteTransactionBegin(struct MultiConnection *connection)
 {
 	RemoteTransaction *transaction = &connection->remoteTransaction;
+	StringInfo beginAndSetDistributedTransactionId = makeStringInfo();
+	DistributedTransactionId *distributedTransactionId = NULL;
+	ListCell *subIdCell = NULL;
+	List *activeSubXacts = NIL;
 
 	Assert(transaction->transactionState == REMOTE_TRANS_INVALID);
 
@@ -51,8 +72,34 @@ StartRemoteTransactionBegin(struct MultiConnection *connection)
 	 * side might have been changed, and that would cause problematic
 	 * behaviour.
 	 */
-	if (!SendRemoteCommand(connection,
-						   "BEGIN TRANSACTION ISOLATION LEVEL READ COMMITTED"))
+	appendStringInfoString(beginAndSetDistributedTransactionId,
+						   "BEGIN TRANSACTION ISOLATION LEVEL READ COMMITTED;");
+
+	/*
+	 * Append BEGIN and assign_distributed_transaction_id() statements into a single command
+	 * and send both in one step. The reason is purely performance, we don't want
+	 * seperate roundtrips for these two statements.
+	 */
+	distributedTransactionId = GetCurrentDistributedTransactionId();
+	appendStringInfo(beginAndSetDistributedTransactionId,
+					 "SELECT assign_distributed_transaction_id(%d, %ld, '%s');",
+					 distributedTransactionId->initiatorNodeIdentifier,
+					 distributedTransactionId->transactionNumber,
+					 timestamptz_to_str(distributedTransactionId->timestamp));
+
+	/* append in-progress savepoints for this transaction */
+	activeSubXacts = ActiveSubXacts();
+	transaction->lastSuccessfulSubXact = TopSubTransactionId;
+	transaction->lastQueuedSubXact = TopSubTransactionId;
+	foreach(subIdCell, activeSubXacts)
+	{
+		SubTransactionId subId = lfirst_int(subIdCell);
+		appendStringInfo(beginAndSetDistributedTransactionId,
+						 "SAVEPOINT savepoint_%u;", subId);
+		transaction->lastQueuedSubXact = subId;
+	}
+
+	if (!SendRemoteCommand(connection, beginAndSetDistributedTransactionId->data))
 	{
 		ReportConnectionError(connection, WARNING);
 		MarkRemoteTransactionFailed(connection, true);
@@ -82,6 +129,7 @@ FinishRemoteTransactionBegin(struct MultiConnection *connection)
 	else
 	{
 		transaction->transactionState = REMOTE_TRANS_STARTED;
+		transaction->lastSuccessfulSubXact = transaction->lastQueuedSubXact;
 	}
 
 	PQclear(result);
@@ -105,6 +153,33 @@ RemoteTransactionBegin(struct MultiConnection *connection)
 }
 
 
+/*
+ * RemoteTransactionListBegin sends BEGIN over all connections in the
+ * given connection list and waits for all of them to finish.
+ */
+void
+RemoteTransactionListBegin(List *connectionList)
+{
+	ListCell *connectionCell = NULL;
+
+	/* send BEGIN to all nodes */
+	foreach(connectionCell, connectionList)
+	{
+		MultiConnection *connection = (MultiConnection *) lfirst(connectionCell);
+
+		StartRemoteTransactionBegin(connection);
+	}
+
+	/* wait for BEGIN to finish on all nodes */
+	foreach(connectionCell, connectionList)
+	{
+		MultiConnection *connection = (MultiConnection *) lfirst(connectionCell);
+
+		FinishRemoteTransactionBegin(connection);
+	}
+}
+
+
 /*
  * StartRemoteTransactionCommit initiates transaction commit in a non-blocking
  * manner.  If the transaction is in a failed state, it'll instead get rolled
@@ -276,16 +351,21 @@ StartRemoteTransactionAbort(MultiConnection *connection)
 	Assert(transaction->transactionState != REMOTE_TRANS_INVALID);
 
 	/*
-	 * Clear previous results, so we have a better chance to send
-	 * ROLLBACK [PREPARED];
+	 * Clear previous results, so we have a better chance to send ROLLBACK
+	 * [PREPARED]. If we've previously sent a PREPARE TRANSACTION, we always
+	 * want to wait for that result, as that shouldn't take long and will
+	 * reserve resources.  But if there's another query running, we don't want
+	 * to wait, because a longrunning statement may be running, force it to be
+	 * killed in that case.
 	 */
-	ForgetResults(connection);
-
 	if (transaction->transactionState == REMOTE_TRANS_PREPARING ||
 		transaction->transactionState == REMOTE_TRANS_PREPARED)
 	{
 		StringInfoData command;
 
+		/* await PREPARE TRANSACTION results, closing the connection would leave it dangling */
+		ForgetResults(connection);
+
 		initStringInfo(&command);
 		appendStringInfo(&command, "ROLLBACK PREPARED '%s'",
 						 transaction->preparedName);
@@ -304,6 +384,14 @@ StartRemoteTransactionAbort(MultiConnection *connection)
 	}
 	else
 	{
+		if (!NonblockingForgetResults(connection))
+		{
+			ShutdownConnection(connection);
+
+			/* FinishRemoteTransactionAbort will emit warning */
+			return;
+		}
+
 		if (!SendRemoteCommand(connection, "ROLLBACK"))
 		{
 			/* no point in reporting a likely redundant message */
@@ -336,16 +424,16 @@ FinishRemoteTransactionAbort(MultiConnection *connection)
 		ReportResultError(connection, result, WARNING);
 		MarkRemoteTransactionFailed(connection, dontRaiseErrors);
 
-		if (transaction->transactionState == REMOTE_TRANS_1PC_ABORTING)
+		if (transaction->transactionState == REMOTE_TRANS_2PC_ABORTING)
 		{
-			ereport(WARNING,
-					(errmsg("failed to abort 2PC transaction \"%s\" on %s:%d",
-							transaction->preparedName, connection->hostname,
-							connection->port)));
+			WarnAboutLeakedPreparedTransaction(connection, isNotCommit);
 		}
 		else
 		{
-			WarnAboutLeakedPreparedTransaction(connection, isNotCommit);
+			ereport(WARNING,
+					(errmsg("failed to abort 1PC transaction \"%s\" on %s:%d",
+							transaction->preparedName, connection->hostname,
+							connection->port)));
 		}
 	}
 
@@ -487,6 +575,7 @@ void
 RemoteTransactionsBeginIfNecessary(List *connectionList)
 {
 	ListCell *connectionCell = NULL;
+	bool raiseInterrupts = true;
 
 	/*
 	 * Don't do anything if not in a coordinated transaction. That allows the
@@ -520,7 +609,8 @@ RemoteTransactionsBeginIfNecessary(List *connectionList)
 		StartRemoteTransactionBegin(connection);
 	}
 
-	/* XXX: Should perform network IO for all connections in a non-blocking manner */
+	raiseInterrupts = true;
+	WaitForAllConnections(connectionList, raiseInterrupts);
 
 	/* get result of all the BEGINs */
 	foreach(connectionCell, connectionList)
@@ -627,6 +717,8 @@ void
 CoordinatedRemoteTransactionsPrepare(void)
 {
 	dlist_iter iter;
+	bool raiseInterrupts = false;
+	List *connectionList = NIL;
 
 	/* issue PREPARE TRANSACTION; to all relevant remote nodes */
 
@@ -646,9 +738,11 @@ CoordinatedRemoteTransactionsPrepare(void)
 		}
 
 		StartRemoteTransactionPrepare(connection);
+		connectionList = lappend(connectionList, connection);
 	}
 
-	/* XXX: Should perform network IO for all connections in a non-blocking manner */
+	raiseInterrupts = true;
+	WaitForAllConnections(connectionList, raiseInterrupts);
 
 	/* Wait for result */
 	dlist_foreach(iter, &InProgressTransactions)
@@ -681,6 +775,8 @@ void
 CoordinatedRemoteTransactionsCommit(void)
 {
 	dlist_iter iter;
+	List *connectionList = NIL;
+	bool raiseInterrupts = false;
 
 	/*
 	 * Before starting to commit on any of the nodes - after which we can't
@@ -704,15 +800,18 @@ CoordinatedRemoteTransactionsCommit(void)
 		if (transaction->transactionState == REMOTE_TRANS_INVALID ||
 			transaction->transactionState == REMOTE_TRANS_1PC_COMMITTING ||
 			transaction->transactionState == REMOTE_TRANS_2PC_COMMITTING ||
-			transaction->transactionState == REMOTE_TRANS_COMMITTED)
+			transaction->transactionState == REMOTE_TRANS_COMMITTED ||
+			transaction->transactionState == REMOTE_TRANS_ABORTED)
 		{
 			continue;
 		}
 
 		StartRemoteTransactionCommit(connection);
+		connectionList = lappend(connectionList, connection);
 	}
 
-	/* XXX: Should perform network IO for all connections in a non-blocking manner */
+	raiseInterrupts = false;
+	WaitForAllConnections(connectionList, raiseInterrupts);
 
 	/* wait for the replies to the commands to come in */
 	dlist_foreach(iter, &InProgressTransactions)
@@ -746,6 +845,8 @@ void
 CoordinatedRemoteTransactionsAbort(void)
 {
 	dlist_iter iter;
+	List *connectionList = NIL;
+	bool raiseInterrupts = false;
 
 	/* asynchronously send ROLLBACK [PREPARED] */
 	dlist_foreach(iter, &InProgressTransactions)
@@ -763,9 +864,11 @@ CoordinatedRemoteTransactionsAbort(void)
 		}
 
 		StartRemoteTransactionAbort(connection);
+		connectionList = lappend(connectionList, connection);
 	}
 
-	/* XXX: Should perform network IO for all connections in a non-blocking manner */
+	raiseInterrupts = false;
+	WaitForAllConnections(connectionList, raiseInterrupts);
 
 	/* and wait for the results */
 	dlist_foreach(iter, &InProgressTransactions)
@@ -785,6 +888,293 @@ CoordinatedRemoteTransactionsAbort(void)
 }
 
 
+/*
+ * CoordinatedRemoteTransactionsSavepointBegin sends the SAVEPOINT command for
+ * the given sub-transaction id to all connections participating in the current
+ * transaction.
+ */
+void
+CoordinatedRemoteTransactionsSavepointBegin(SubTransactionId subId)
+{
+	dlist_iter iter;
+	const bool raiseInterrupts = true;
+	List *connectionList = NIL;
+
+	/* asynchronously send SAVEPOINT */
+	dlist_foreach(iter, &InProgressTransactions)
+	{
+		MultiConnection *connection = dlist_container(MultiConnection, transactionNode,
+													  iter.cur);
+		RemoteTransaction *transaction = &connection->remoteTransaction;
+		if (transaction->transactionFailed)
+		{
+			continue;
+		}
+
+		StartRemoteTransactionSavepointBegin(connection, subId);
+		connectionList = lappend(connectionList, connection);
+	}
+
+	WaitForAllConnections(connectionList, raiseInterrupts);
+
+	/* and wait for the results */
+	dlist_foreach(iter, &InProgressTransactions)
+	{
+		MultiConnection *connection = dlist_container(MultiConnection, transactionNode,
+													  iter.cur);
+		RemoteTransaction *transaction = &connection->remoteTransaction;
+		if (transaction->transactionFailed)
+		{
+			continue;
+		}
+
+		FinishRemoteTransactionSavepointBegin(connection, subId);
+
+		if (!transaction->transactionFailed)
+		{
+			transaction->lastSuccessfulSubXact = subId;
+		}
+	}
+}
+
+
+/*
+ * CoordinatedRemoteTransactionsSavepointRelease sends the RELEASE SAVEPOINT
+ * command for the given sub-transaction id to all connections participating in
+ * the current transaction.
+ */
+void
+CoordinatedRemoteTransactionsSavepointRelease(SubTransactionId subId)
+{
+	dlist_iter iter;
+	const bool raiseInterrupts = true;
+	List *connectionList = NIL;
+
+	/* asynchronously send RELEASE SAVEPOINT */
+	dlist_foreach(iter, &InProgressTransactions)
+	{
+		MultiConnection *connection = dlist_container(MultiConnection, transactionNode,
+													  iter.cur);
+		RemoteTransaction *transaction = &connection->remoteTransaction;
+		if (transaction->transactionFailed)
+		{
+			continue;
+		}
+
+		StartRemoteTransactionSavepointRelease(connection, subId);
+		connectionList = lappend(connectionList, connection);
+	}
+
+	WaitForAllConnections(connectionList, raiseInterrupts);
+
+	/* and wait for the results */
+	dlist_foreach(iter, &InProgressTransactions)
+	{
+		MultiConnection *connection = dlist_container(MultiConnection, transactionNode,
+													  iter.cur);
+		RemoteTransaction *transaction = &connection->remoteTransaction;
+		if (transaction->transactionFailed)
+		{
+			continue;
+		}
+
+		FinishRemoteTransactionSavepointRelease(connection, subId);
+	}
+}
+
+
+/*
+ * CoordinatedRemoteTransactionsSavepointRollback sends the ROLLBACK TO SAVEPOINT
+ * command for the given sub-transaction id to all connections participating in
+ * the current transaction.
+ */
+void
+CoordinatedRemoteTransactionsSavepointRollback(SubTransactionId subId)
+{
+	dlist_iter iter;
+	const bool dontRaiseInterrupts = false;
+	List *connectionList = NIL;
+
+	/* asynchronously send ROLLBACK TO SAVEPOINT */
+	dlist_foreach(iter, &InProgressTransactions)
+	{
+		MultiConnection *connection = dlist_container(MultiConnection, transactionNode,
+													  iter.cur);
+		RemoteTransaction *transaction = &connection->remoteTransaction;
+		if (transaction->transactionFailed)
+		{
+			if (transaction->lastSuccessfulSubXact <= subId)
+			{
+				transaction->transactionRecovering = true;
+
+				/*
+				 * Clear the results of the failed query so we can send the ROLLBACK
+				 * TO SAVEPOINT command for a savepoint that can recover the transaction
+				 * from failure.
+				 */
+				ForgetResults(connection);
+			}
+			else
+			{
+				continue;
+			}
+		}
+		StartRemoteTransactionSavepointRollback(connection, subId);
+		connectionList = lappend(connectionList, connection);
+	}
+
+	WaitForAllConnections(connectionList, dontRaiseInterrupts);
+
+	/* and wait for the results */
+	dlist_foreach(iter, &InProgressTransactions)
+	{
+		MultiConnection *connection = dlist_container(MultiConnection, transactionNode,
+													  iter.cur);
+		RemoteTransaction *transaction = &connection->remoteTransaction;
+		if (transaction->transactionFailed && !transaction->transactionRecovering)
+		{
+			continue;
+		}
+
+		FinishRemoteTransactionSavepointRollback(connection, subId);
+	}
+}
+
+
+/*
+ * StartRemoteTransactionSavepointBegin initiates SAVEPOINT command for the given
+ * subtransaction id in a non-blocking manner.
+ */
+static void
+StartRemoteTransactionSavepointBegin(MultiConnection *connection, SubTransactionId subId)
+{
+	const bool raiseErrors = true;
+	StringInfo savepointCommand = makeStringInfo();
+	appendStringInfo(savepointCommand, "SAVEPOINT savepoint_%u", subId);
+
+	if (!SendRemoteCommand(connection, savepointCommand->data))
+	{
+		ReportConnectionError(connection, WARNING);
+		MarkRemoteTransactionFailed(connection, raiseErrors);
+	}
+}
+
+
+/*
+ * FinishRemoteTransactionSavepointBegin finishes the work
+ * StartRemoteTransactionSavepointBegin initiated. It blocks if necessary (i.e.
+ * if PQisBusy() would return true).
+ */
+static void
+FinishRemoteTransactionSavepointBegin(MultiConnection *connection, SubTransactionId subId)
+{
+	const bool raiseErrors = true;
+	PGresult *result = GetRemoteCommandResult(connection, raiseErrors);
+	if (!IsResponseOK(result))
+	{
+		ReportResultError(connection, result, WARNING);
+		MarkRemoteTransactionFailed(connection, raiseErrors);
+	}
+
+	PQclear(result);
+	ForgetResults(connection);
+}
+
+
+/*
+ * StartRemoteTransactionSavepointRelease initiates RELEASE SAVEPOINT command for
+ * the given subtransaction id in a non-blocking manner.
+ */
+static void
+StartRemoteTransactionSavepointRelease(MultiConnection *connection,
+									   SubTransactionId subId)
+{
+	const bool raiseErrors = true;
+	StringInfo savepointCommand = makeStringInfo();
+	appendStringInfo(savepointCommand, "RELEASE SAVEPOINT savepoint_%u", subId);
+
+	if (!SendRemoteCommand(connection, savepointCommand->data))
+	{
+		ReportConnectionError(connection, WARNING);
+		MarkRemoteTransactionFailed(connection, raiseErrors);
+	}
+}
+
+
+/*
+ * FinishRemoteTransactionSavepointRelease finishes the work
+ * StartRemoteTransactionSavepointRelease initiated. It blocks if necessary (i.e.
+ * if PQisBusy() would return true).
+ */
+static void
+FinishRemoteTransactionSavepointRelease(MultiConnection *connection,
+										SubTransactionId subId)
+{
+	const bool raiseErrors = true;
+	PGresult *result = GetRemoteCommandResult(connection, raiseErrors);
+	if (!IsResponseOK(result))
+	{
+		ReportResultError(connection, result, WARNING);
+		MarkRemoteTransactionFailed(connection, raiseErrors);
+	}
+
+	PQclear(result);
+	ForgetResults(connection);
+}
+
+
+/*
+ * StartRemoteTransactionSavepointRollback initiates ROLLBACK TO SAVEPOINT command
+ * for the given subtransaction id in a non-blocking manner.
+ */
+static void
+StartRemoteTransactionSavepointRollback(MultiConnection *connection,
+										SubTransactionId subId)
+{
+	const bool dontRaiseErrors = false;
+	StringInfo savepointCommand = makeStringInfo();
+	appendStringInfo(savepointCommand, "ROLLBACK TO SAVEPOINT savepoint_%u", subId);
+
+	if (!SendRemoteCommand(connection, savepointCommand->data))
+	{
+		ReportConnectionError(connection, WARNING);
+		MarkRemoteTransactionFailed(connection, dontRaiseErrors);
+	}
+}
+
+
+/*
+ * FinishRemoteTransactionSavepointRollback finishes the work
+ * StartRemoteTransactionSavepointRollback initiated. It blocks if necessary (i.e.
+ * if PQisBusy() would return true). It also recovers the transaction from failure
+ * if transaction is recovering and the rollback command succeeds.
+ */
+static void
+FinishRemoteTransactionSavepointRollback(MultiConnection *connection, SubTransactionId
+										 subId)
+{
+	const bool dontRaiseErrors = false;
+	RemoteTransaction *transaction = &connection->remoteTransaction;
+
+	PGresult *result = GetRemoteCommandResult(connection, dontRaiseErrors);
+	if (!IsResponseOK(result))
+	{
+		ReportResultError(connection, result, WARNING);
+		MarkRemoteTransactionFailed(connection, dontRaiseErrors);
+	}
+
+	/* ROLLBACK TO SAVEPOINT succeeded, check if it recovers the transaction */
+	else if (transaction->transactionRecovering)
+	{
+		transaction->transactionFailed = false;
+		transaction->transactionRecovering = false;
+	}
+
+	PQclear(result);
+	ForgetResults(connection);
+}
+
+
 /*
  * CheckTransactionHealth checks if any of the participating transactions in a
  * coordinated transaction failed, and what consequence that should have.
@@ -823,11 +1213,27 @@ CheckTransactionHealth(void)
 
 
 /*
- * Assign2PCIdentifier compute the 2PC transaction name to use for a
- * transaction.
+ * Assign2PCIdentifier computes the 2PC transaction name to use for a
+ * transaction. Every prepared transaction should get a new name, i.e. this
+ * function will need to be called again.
  *
- * Every 2PC transaction should get a new name, i.e. this function will need
- * to be called again.
+ * The format of the name is:
+ *
+ * citus_<source group>_<pid>_<distributed transaction number>_<connection number>
+ *
+ * (at most 5+1+10+1+10+20+1+10 = 58 characters, while limit is 64)
+ *
+ * The source group is used to distinguish 2PCs started by different
+ * coordinators. A coordinator will only attempt to recover its own 2PCs.
+ *
+ * The pid is used to distinguish different processes on the coordinator, mainly
+ * to provide some entropy across restarts.
+ *
+ * The distributed transaction number is used to distinguish different
+ * transactions originating from the same node (since restart).
+ *
+ * The connection number is used to distinguish connections made to a node
+ * within the same transaction.
  *
  * NB: we rely on the fact that we don't need to do full escaping on the names
  * generated here.
@@ -835,10 +1241,16 @@ CheckTransactionHealth(void)
 static void
 Assign2PCIdentifier(MultiConnection *connection)
 {
-	static uint64 sequence = 0;
+	/* local sequence number used to distinguish different connections */
+	static uint32 connectionNumber = 0;
+
+	/* transaction identifier that is unique across processes */
+	uint64 transactionNumber = CurrentDistributedTransactionNumber();
+
+	/* print all numbers as unsigned to guarantee no minus symbols appear in the name */
 	snprintf(connection->remoteTransaction.preparedName, NAMEDATALEN,
-			 "citus_%d_%d_"UINT64_FORMAT, GetLocalGroupId(),
-			 MyProcPid, sequence++);
+			 "citus_%u_%u_"UINT64_FORMAT "_%u", GetLocalGroupId(), MyProcPid,
+			 transactionNumber, connectionNumber++);
 }
 
 
