@@ -100,12 +100,12 @@ LogTransactionRecord(int groupId, char *transactionName)
 	tupleDescriptor = RelationGetDescr(pgDistTransaction);
 	heapTuple = heap_form_tuple(tupleDescriptor, values, isNulls);
 
-	simple_heap_insert(pgDistTransaction, heapTuple);
-	CatalogUpdateIndexes(pgDistTransaction, heapTuple);
+	CatalogTupleInsert(pgDistTransaction, heapTuple);
+
 	CommandCounterIncrement();
 
 	/* close relation and invalidate previous cache entry */
-	heap_close(pgDistTransaction, RowExclusiveLock);
+	heap_close(pgDistTransaction, NoLock);
 }
 
 
@@ -127,7 +127,7 @@ RecoverPreparedTransactions(void)
 	 */
 	LockRelationOid(DistTransactionRelationId(), ExclusiveLock);
 
-	workerList = ActiveWorkerNodeList();
+	workerList = ActivePrimaryNodeList();
 
 	foreach(workerNodeCell, workerList)
 	{
