@@ -14,6 +14,7 @@
 
 #include "postgres.h"
 #include "funcapi.h"
+#include "libpq-fe.h"
 #include "miscadmin.h"
 #include <unistd.h>
 #include <sys/stat.h>
@@ -26,12 +27,15 @@
 #include "commands/extension.h"
 #include "commands/sequence.h"
 #include "distributed/citus_ruleutils.h"
+#include "distributed/connection_management.h"
 #include "distributed/master_protocol.h"
 #include "distributed/metadata_cache.h"
 #include "distributed/multi_client_executor.h"
 #include "distributed/multi_logical_optimizer.h"
 #include "distributed/multi_server_executor.h"
+#include "distributed/multi_utility.h"
 #include "distributed/relay_utility.h"
+#include "distributed/remote_commands.h"
 #include "distributed/resource_lock.h"
 #include "distributed/task_tracker.h"
 #include "distributed/worker_protocol.h"
@@ -41,6 +45,10 @@
 #include "tcop/utility.h"
 #include "utils/builtins.h"
 #include "utils/lsyscache.h"
+#if (PG_VERSION_NUM >= 100000)
+#include "utils/regproc.h"
+#include "utils/varlena.h"
+#endif
 
 
 /* Config variable managed via guc.c */
@@ -425,8 +433,8 @@ worker_apply_shard_ddl_command(PG_FUNCTION_ARGS)
 
 	/* extend names in ddl command and apply extended command */
 	RelayEventExtendNames(ddlCommandNode, schemaName, shardId);
-	ProcessUtility(ddlCommandNode, ddlCommand, PROCESS_UTILITY_TOPLEVEL,
-				   NULL, None_Receiver, NULL);
+	CitusProcessUtility(ddlCommandNode, ddlCommand, PROCESS_UTILITY_TOPLEVEL, NULL,
+						None_Receiver, NULL);
 
 	PG_RETURN_VOID();
 }
@@ -457,8 +465,8 @@ worker_apply_inter_shard_ddl_command(PG_FUNCTION_ARGS)
 	RelayEventExtendNamesForInterShardCommands(ddlCommandNode, leftShardId,
 											   leftShardSchemaName, rightShardId,
 											   rightShardSchemaName);
-	ProcessUtility(ddlCommandNode, ddlCommand, PROCESS_UTILITY_TOPLEVEL, NULL,
-				   None_Receiver, NULL);
+	CitusProcessUtility(ddlCommandNode, ddlCommand, PROCESS_UTILITY_TOPLEVEL, NULL,
+						None_Receiver, NULL);
 
 	PG_RETURN_VOID();
 }
@@ -493,8 +501,9 @@ worker_apply_sequence_command(PG_FUNCTION_ARGS)
 	}
 
 	/* run the CREATE SEQUENCE command */
-	ProcessUtility(commandNode, commandString, PROCESS_UTILITY_TOPLEVEL,
-				   NULL, None_Receiver, NULL);
+	CitusProcessUtility(commandNode, commandString, PROCESS_UTILITY_TOPLEVEL, NULL,
+						None_Receiver, NULL);
+	CommandCounterIncrement();
 
 	createSequenceStatement = (CreateSeqStmt *) commandNode;
 
@@ -681,7 +690,7 @@ LocalTableSize(Oid relationId)
 	Datum relationIdDatum = ObjectIdGetDatum(relationId);
 
 	relationType = get_rel_relkind(relationId);
-	if (relationType == RELKIND_RELATION)
+	if (RegularTable(relationId))
 	{
 		Datum tableSizeDatum = DirectFunctionCall1(pg_table_size, relationIdDatum);
 
@@ -847,8 +856,8 @@ FetchRegularTable(const char *nodeName, uint32 nodePort, const char *tableName)
 		StringInfo ddlCommand = (StringInfo) lfirst(ddlCommandCell);
 		Node *ddlCommandNode = ParseTreeNode(ddlCommand->data);
 
-		ProcessUtility(ddlCommandNode, ddlCommand->data, PROCESS_UTILITY_TOPLEVEL,
-					   NULL, None_Receiver, NULL);
+		CitusProcessUtility(ddlCommandNode, ddlCommand->data, PROCESS_UTILITY_TOPLEVEL,
+							NULL, None_Receiver, NULL);
 		CommandCounterIncrement();
 	}
 
@@ -866,8 +875,8 @@ FetchRegularTable(const char *nodeName, uint32 nodePort, const char *tableName)
 	queryString = makeStringInfo();
 	appendStringInfo(queryString, COPY_IN_COMMAND, tableName, localFilePath->data);
 
-	ProcessUtility((Node *) localCopyCommand, queryString->data,
-				   PROCESS_UTILITY_TOPLEVEL, NULL, None_Receiver, NULL);
+	CitusProcessUtility((Node *) localCopyCommand, queryString->data,
+						PROCESS_UTILITY_TOPLEVEL, NULL, None_Receiver, NULL);
 
 	/* finally delete the temporary file we created */
 	DeleteFile(localFilePath->data);
@@ -941,8 +950,8 @@ FetchForeignTable(const char *nodeName, uint32 nodePort, const char *tableName)
 		StringInfo ddlCommand = (StringInfo) lfirst(ddlCommandCell);
 		Node *ddlCommandNode = ParseTreeNode(ddlCommand->data);
 
-		ProcessUtility(ddlCommandNode, ddlCommand->data, PROCESS_UTILITY_TOPLEVEL,
-					   NULL, None_Receiver, NULL);
+		CitusProcessUtility(ddlCommandNode, ddlCommand->data, PROCESS_UTILITY_TOPLEVEL,
+							NULL, None_Receiver, NULL);
 		CommandCounterIncrement();
 	}
 
@@ -960,11 +969,17 @@ RemoteTableOwner(const char *nodeName, uint32 nodePort, const char *tableName)
 	List *ownerList = NIL;
 	StringInfo queryString = NULL;
 	StringInfo relationOwner;
+	MultiConnection *connection = NULL;
+	uint32 connectionFlag = FORCE_NEW_CONNECTION;
+	PGresult *result = NULL;
 
 	queryString = makeStringInfo();
 	appendStringInfo(queryString, GET_TABLE_OWNER, tableName);
+	connection = GetNodeConnection(connectionFlag, nodeName, nodePort);
 
-	ownerList = ExecuteRemoteQuery(nodeName, nodePort, NULL, queryString);
+	ExecuteOptionalRemoteCommand(connection, queryString->data, &result);
+
+	ownerList = ReadFirstColumnAsText(result);
 	if (list_length(ownerList) != 1)
 	{
 		return NULL;
@@ -986,11 +1001,20 @@ TableDDLCommandList(const char *nodeName, uint32 nodePort, const char *tableName
 {
 	List *ddlCommandList = NIL;
 	StringInfo queryString = NULL;
+	MultiConnection *connection = NULL;
+	PGresult *result = NULL;
+	uint32 connectionFlag = FORCE_NEW_CONNECTION;
 
 	queryString = makeStringInfo();
 	appendStringInfo(queryString, GET_TABLE_DDL_EVENTS, tableName);
+	connection = GetNodeConnection(connectionFlag, nodeName, nodePort);
+
+	ExecuteOptionalRemoteCommand(connection, queryString->data, &result);
+	ddlCommandList = ReadFirstColumnAsText(result);
+
+	ForgetResults(connection);
+	CloseConnection(connection);
 
-	ddlCommandList = ExecuteRemoteQuery(nodeName, nodePort, NULL, queryString);
 	return ddlCommandList;
 }
 
@@ -1006,11 +1030,17 @@ ForeignFilePath(const char *nodeName, uint32 nodePort, const char *tableName)
 	List *foreignPathList = NIL;
 	StringInfo foreignPathCommand = NULL;
 	StringInfo foreignPath = NULL;
+	MultiConnection *connection = NULL;
+	PGresult *result = NULL;
+	int connectionFlag = FORCE_NEW_CONNECTION;
 
 	foreignPathCommand = makeStringInfo();
 	appendStringInfo(foreignPathCommand, FOREIGN_FILE_PATH_COMMAND, tableName);
+	connection = GetNodeConnection(connectionFlag, nodeName, nodePort);
 
-	foreignPathList = ExecuteRemoteQuery(nodeName, nodePort, NULL, foreignPathCommand);
+	ExecuteOptionalRemoteCommand(connection, foreignPathCommand->data, &result);
+
+	foreignPathList = ReadFirstColumnAsText(result);
 	if (foreignPathList != NIL)
 	{
 		foreignPath = (StringInfo) linitial(foreignPathList);
@@ -1101,61 +1131,18 @@ ExecuteRemoteQuery(const char *nodeName, uint32 nodePort, char *runAsUser,
 
 
 /*
- * ExecuteRemoteCommand executes the given SQL command. This command could be an
- * Insert, Update, or Delete statement, or a utility command that returns
- * nothing. If query is successfuly executed, the function returns true.
- * Otherwise, it returns false.
+ * Parses the given DDL command, and returns the tree node for parsed command.
  */
-bool
-ExecuteRemoteCommand(const char *nodeName, uint32 nodePort, StringInfo queryString)
+Node *
+ParseTreeNode(const char *ddlCommand)
 {
-	char *nodeDatabase = get_database_name(MyDatabaseId);
-	int32 connectionId = -1;
-	QueryStatus queryStatus = CLIENT_INVALID_QUERY;
-	bool querySent = false;
-	bool queryReady = false;
-	bool queryDone = false;
+	Node *parseTreeNode = ParseTreeRawStmt(ddlCommand);
 
-	connectionId = MultiClientConnect(nodeName, nodePort, nodeDatabase, NULL);
-	if (connectionId == INVALID_CONNECTION_ID)
-	{
-		return false;
-	}
+#if (PG_VERSION_NUM >= 100000)
+	parseTreeNode = ((RawStmt *) parseTreeNode)->stmt;
+#endif
 
-	querySent = MultiClientSendQuery(connectionId, queryString->data);
-	if (!querySent)
-	{
-		MultiClientDisconnect(connectionId);
-		return false;
-	}
-
-	while (!queryReady)
-	{
-		ResultStatus resultStatus = MultiClientResultStatus(connectionId);
-		if (resultStatus == CLIENT_RESULT_READY)
-		{
-			queryReady = true;
-		}
-		else if (resultStatus == CLIENT_RESULT_BUSY)
-		{
-			long sleepIntervalPerCycle = RemoteTaskCheckInterval * 1000L;
-			pg_usleep(sleepIntervalPerCycle);
-		}
-		else
-		{
-			MultiClientDisconnect(connectionId);
-			return false;
-		}
-	}
-
-	queryStatus = MultiClientQueryStatus(connectionId);
-	if (queryStatus == CLIENT_QUERY_DONE)
-	{
-		queryDone = true;
-	}
-
-	MultiClientDisconnect(connectionId);
-	return queryDone;
+	return parseTreeNode;
 }
 
 
@@ -1163,7 +1150,7 @@ ExecuteRemoteCommand(const char *nodeName, uint32 nodePort, StringInfo queryStri
  * Parses the given DDL command, and returns the tree node for parsed command.
  */
 Node *
-ParseTreeNode(const char *ddlCommand)
+ParseTreeRawStmt(const char *ddlCommand)
 {
 	Node *parseTreeNode = NULL;
 	List *parseTreeList = NULL;
@@ -1271,8 +1258,8 @@ worker_append_table_to_shard(PG_FUNCTION_ARGS)
 	appendStringInfo(queryString, COPY_IN_COMMAND, shardQualifiedName,
 					 localFilePath->data);
 
-	ProcessUtility((Node *) localCopyCommand, queryString->data,
-				   PROCESS_UTILITY_TOPLEVEL, NULL, None_Receiver, NULL);
+	CitusProcessUtility((Node *) localCopyCommand, queryString->data,
+						PROCESS_UTILITY_TOPLEVEL, NULL, None_Receiver, NULL);
 
 	/* finally delete the temporary file we created */
 	DeleteFile(localFilePath->data);
@@ -1333,6 +1320,14 @@ AlterSequenceMinMax(Oid sequenceId, char *schemaName, char *sequenceName)
 	Form_pg_sequence sequenceData = pg_get_sequencedef(sequenceId);
 	int64 startValue = 0;
 	int64 maxValue = 0;
+#if (PG_VERSION_NUM >= 100000)
+	int64 sequenceMaxValue = sequenceData->seqmax;
+	int64 sequenceMinValue = sequenceData->seqmin;
+#else
+	int64 sequenceMaxValue = sequenceData->max_value;
+	int64 sequenceMinValue = sequenceData->min_value;
+#endif
+
 
 	/* calculate min/max values that the sequence can generate in this worker */
 	startValue = (((int64) GetLocalGroupId()) << 48) + 1;
@@ -1343,7 +1338,7 @@ AlterSequenceMinMax(Oid sequenceId, char *schemaName, char *sequenceName)
 	 * their correct values. This happens when the sequence has been created
 	 * during shard, before the current worker having the metadata.
 	 */
-	if (sequenceData->min_value != startValue || sequenceData->max_value != maxValue)
+	if (sequenceMinValue != startValue || sequenceMaxValue != maxValue)
 	{
 		StringInfo startNumericString = makeStringInfo();
 		StringInfo maxNumericString = makeStringInfo();
@@ -1371,8 +1366,8 @@ AlterSequenceMinMax(Oid sequenceId, char *schemaName, char *sequenceName)
 		SetDefElemArg(alterSequenceStatement, "restart", startFloatArg);
 
 		/* since the command is an AlterSeqStmt, a dummy command string works fine */
-		ProcessUtility((Node *) alterSequenceStatement, dummyString,
-					   PROCESS_UTILITY_TOPLEVEL, NULL, None_Receiver, NULL);
+		CitusProcessUtility((Node *) alterSequenceStatement, dummyString,
+							PROCESS_UTILITY_TOPLEVEL, NULL, None_Receiver, NULL);
 	}
 }
 
@@ -1402,6 +1397,11 @@ SetDefElemArg(AlterSeqStmt *statement, const char *name, Node *arg)
 		}
 	}
 
+#if (PG_VERSION_NUM >= 100000)
+	defElem = makeDefElem((char *) name, arg, -1);
+#else
 	defElem = makeDefElem((char *) name, arg);
+#endif
+
 	statement->options = lappend(statement->options, defElem);
 }
