@@ -21,7 +21,6 @@
 
 #define MAX_TASK_EXECUTION_FAILURES 3 /* allowed failure count for one task */
 #define MAX_TRACKER_FAILURE_COUNT 3   /* allowed failure count for one tracker */
-#define REMOTE_NODE_CONNECT_TIMEOUT 4000 /* async connect timeout in ms */
 #define RESERVED_FD_COUNT 64           /* file descriptors unavailable to executor */
 
 /* copy out query results */
@@ -32,11 +31,13 @@
 
 /* Task tracker executor related defines */
 #define TASK_ASSIGNMENT_QUERY "SELECT task_tracker_assign_task \
- ("UINT64_FORMAT ", %u, %s)"
-#define TASK_STATUS_QUERY "SELECT task_tracker_task_status("UINT64_FORMAT ", %u)"
+ ("UINT64_FORMAT ", %u, %s);"
+#define TASK_STATUS_QUERY "SELECT task_tracker_task_status("UINT64_FORMAT ", %u);"
 #define JOB_CLEANUP_QUERY "SELECT task_tracker_cleanup_job("UINT64_FORMAT ")"
 #define JOB_CLEANUP_TASK_ID INT_MAX
 
+#define MULTI_TASK_QUERY_INFO_OFF 0  /* do not log multi-task queries */
+
 
 /* Enumeration to track one task's execution status */
 typedef enum
@@ -93,7 +94,8 @@ typedef enum
 	MULTI_EXECUTOR_INVALID_FIRST = 0,
 	MULTI_EXECUTOR_REAL_TIME = 1,
 	MULTI_EXECUTOR_TASK_TRACKER = 2,
-	MULTI_EXECUTOR_ROUTER = 3
+	MULTI_EXECUTOR_ROUTER = 3,
+	MULTI_EXECUTOR_COORDINATOR_INSERT_SELECT = 4
 } MultiExecutorType;
 
 
@@ -114,6 +116,7 @@ typedef enum
  */
 struct TaskExecution
 {
+	CitusNode type;
 	uint64 jobId;
 	uint32 taskId;
 
@@ -164,6 +167,7 @@ typedef struct TaskTracker
 	int32 currentTaskIndex;
 	bool connectionBusy;
 	TrackerTaskState *connectionBusyOnTask;
+	List *connectionBusyOnTaskList;
 } TaskTracker;
 
 
@@ -184,6 +188,7 @@ extern int RemoteTaskCheckInterval;
 extern int MaxAssignTaskBatchSize;
 extern int TaskExecutorType;
 extern bool BinaryMasterCopyFormat;
+extern int MultiTaskQueryLogLevel;
 
 
 /* Function declarations for distributed execution */
