@@ -10,6 +10,34 @@
 ALTER SEQUENCE pg_catalog.pg_dist_shardid_seq RESTART 580000;
 ALTER SEQUENCE pg_catalog.pg_dist_jobid_seq RESTART 580000;
 
+CREATE SCHEMA test;
+
+CREATE OR REPLACE FUNCTION test.maintenance_worker(p_dbname text DEFAULT current_database())
+    RETURNS pg_stat_activity
+    LANGUAGE plpgsql
+AS $$
+DECLARE
+   activity record;
+BEGIN
+    LOOP
+        SELECT * INTO activity FROM pg_stat_activity
+	WHERE application_name = 'Citus Maintenance Daemon' AND datname = p_dbname;
+        IF activity.pid IS NOT NULL THEN
+            RETURN activity;
+        ELSE
+            PERFORM pg_sleep(0.1);
+            PERFORM pg_stat_clear_snapshot();
+        END IF ;
+    END LOOP;
+END;
+$$;
+
+-- check maintenance daemon is started
+SELECT datname,
+    datname = current_database(),
+    usename = (SELECT extowner::regrole::text FROM pg_extension WHERE extname = 'citus')
+FROM test.maintenance_worker();
+
 -- ensure no objects were created outside pg_catalog
 SELECT COUNT(*)
 FROM pg_depend AS pgd,
@@ -18,7 +46,8 @@ FROM pg_depend AS pgd,
 WHERE pgd.refclassid = 'pg_extension'::regclass AND
 	  pgd.refobjid   = pge.oid AND
 	  pge.extname    = 'citus' AND
-	  pgio.schema    NOT IN ('pg_catalog', 'citus');
+	  pgio.schema    NOT IN ('pg_catalog', 'citus', 'test');
+
 
 -- DROP EXTENSION pre-created by the regression suite
 DROP EXTENSION citus;
@@ -81,6 +110,21 @@ ALTER EXTENSION citus UPDATE TO '6.2-1';
 ALTER EXTENSION citus UPDATE TO '6.2-2';
 ALTER EXTENSION citus UPDATE TO '6.2-3';
 ALTER EXTENSION citus UPDATE TO '6.2-4';
+ALTER EXTENSION citus UPDATE TO '7.0-1';
+ALTER EXTENSION citus UPDATE TO '7.0-2';
+ALTER EXTENSION citus UPDATE TO '7.0-3';
+ALTER EXTENSION citus UPDATE TO '7.0-4';
+ALTER EXTENSION citus UPDATE TO '7.0-5';
+ALTER EXTENSION citus UPDATE TO '7.0-6';
+ALTER EXTENSION citus UPDATE TO '7.0-7';
+ALTER EXTENSION citus UPDATE TO '7.0-8';
+ALTER EXTENSION citus UPDATE TO '7.0-9';
+ALTER EXTENSION citus UPDATE TO '7.0-10';
+ALTER EXTENSION citus UPDATE TO '7.0-11';
+ALTER EXTENSION citus UPDATE TO '7.0-12';
+ALTER EXTENSION citus UPDATE TO '7.0-13';
+ALTER EXTENSION citus UPDATE TO '7.0-14';
+ALTER EXTENSION citus UPDATE TO '7.0-15';
 
 -- show running version
 SHOW citus.version;
@@ -93,7 +137,7 @@ FROM pg_depend AS pgd,
 WHERE pgd.refclassid = 'pg_extension'::regclass AND
 	  pgd.refobjid   = pge.oid AND
 	  pge.extname    = 'citus' AND
-	  pgio.schema    NOT IN ('pg_catalog', 'citus');
+	  pgio.schema    NOT IN ('pg_catalog', 'citus', 'test');
 
 -- see incompatible version errors out
 RESET citus.enable_version_checks;
@@ -102,7 +146,7 @@ CREATE EXTENSION citus VERSION '5.0';
 
 -- Test non-distributed queries work even in version mismatch
 SET citus.enable_version_checks TO 'false';
-CREATE EXTENSION citus VERSION '6.1-17';
+CREATE EXTENSION citus VERSION '6.2-1';
 SET citus.enable_version_checks TO 'true';
 
 -- Test CREATE TABLE
@@ -119,7 +163,7 @@ CREATE TABLE version_mismatch_table(column1 int);
 
 -- Test INSERT
 INSERT INTO version_mismatch_table(column1) VALUES(5);
- 
+
 -- Test SELECT
 SELECT * FROM version_mismatch_table ORDER BY column1;
 
@@ -163,8 +207,6 @@ CREATE EXTENSION citus;
 -- test cache invalidation in workers
 \c - - - :worker_1_port
 
--- this will initialize the cache
-\d
 DROP EXTENSION citus;
 SET citus.enable_version_checks TO 'false';
 CREATE EXTENSION citus VERSION '5.2-4';
@@ -174,3 +216,94 @@ ALTER EXTENSION citus UPDATE;
 
 -- if cache is invalidated succesfull, this \d should work without any problem
 \d
+
+\c - - - :master_port
+
+-- check that maintenance daemon gets (re-)started for the right user
+DROP EXTENSION citus;
+CREATE USER testuser SUPERUSER;
+SET ROLE testuser;
+CREATE EXTENSION citus;
+
+SELECT datname,
+    datname = current_database(),
+    usename = (SELECT extowner::regrole::text FROM pg_extension WHERE extname = 'citus')
+FROM test.maintenance_worker();
+
+-- and recreate as the right owner
+RESET ROLE;
+DROP EXTENSION citus;
+CREATE EXTENSION citus;
+
+
+-- Check that maintenance daemon can also be started in another database
+CREATE DATABASE another;
+\c another
+CREATE EXTENSION citus;
+
+CREATE SCHEMA test;
+
+CREATE OR REPLACE FUNCTION test.maintenance_worker(p_dbname text DEFAULT current_database())
+    RETURNS pg_stat_activity
+    LANGUAGE plpgsql
+AS $$
+DECLARE
+   activity record;
+BEGIN
+    LOOP
+        SELECT * INTO activity FROM pg_stat_activity
+	WHERE application_name = 'Citus Maintenance Daemon' AND datname = p_dbname;
+        IF activity.pid IS NOT NULL THEN
+            RETURN activity;
+        ELSE
+            PERFORM pg_sleep(0.1);
+            PERFORM pg_stat_clear_snapshot();
+        END IF ;
+    END LOOP;
+END;
+$$;
+
+-- see that the deamon started
+SELECT datname,
+    datname = current_database(),
+    usename = (SELECT extowner::regrole::text FROM pg_extension WHERE extname = 'citus')
+FROM test.maintenance_worker();
+
+-- Test that database with active worker can be dropped.
+\c regression
+
+CREATE SCHEMA test_deamon;
+
+-- we create a similar function on the regression database
+-- note that this function checks for the existence of the daemon
+-- when not found, returns true else tries for 5 times and 
+-- returns false
+CREATE OR REPLACE FUNCTION test_deamon.maintenance_deamon_died(p_dbname text)
+    RETURNS boolean
+    LANGUAGE plpgsql
+AS $$
+DECLARE
+   activity record;
+BEGIN
+    PERFORM pg_stat_clear_snapshot();
+    LOOP
+        SELECT * INTO activity FROM pg_stat_activity
+        WHERE application_name = 'Citus Maintenance Daemon' AND datname = p_dbname;
+        IF activity.pid IS NULL THEN
+            RETURN true;
+        ELSE
+            RETURN false;
+        END IF;
+    END LOOP;
+END;
+$$;
+
+-- drop the database and see that the deamon is dead
+DROP DATABASE another;
+SELECT 
+    *
+FROM 
+    test_deamon.maintenance_deamon_died('another');
+
+-- we don't need the schema and the function anymore
+DROP SCHEMA test_deamon CASCADE;
